@@ -49,6 +49,7 @@ killrpath(const char *filename)
    if (0 != elf_find_dynamic_section(fd, &ehdr, &phdr))
    {
      perror("found no dynamic section");
+     elf_close(fd);
      return 1;
    }
 
@@ -56,13 +57,16 @@ killrpath(const char *filename)
    if (dyns == NULL)
      {
        perror ("allocating memory for dynamic section");
+       elf_close(fd);
        return 1;
      }
    memset(dyns, 0, PHDR(p_memsz));
    if (lseek(fd, PHDR(p_offset), SEEK_SET) == -1
-       || read(fd, dyns, PHDR(p_filesz)) != (int)PHDR(p_filesz))
+       || read(fd, dyns, PHDR(p_filesz)) != (ssize_t)PHDR(p_filesz))
      {
        perror ("reading dynamic section");
+       free(dyns);
+       elf_close(fd);
        return 1;
      }
 
@@ -78,19 +82,25 @@ killrpath(const char *filename)
      }
    for (; dynpos < i; dynpos++)
      {
-       if (is_e32())
+       if (is_e32()) {
         ((Elf32_Dyn *)dyns)[dynpos].d_tag = DT_NULL;
-       else
+        ((Elf32_Dyn *)dyns)[dynpos].d_un.d_val = 0x0;
+       } else {
         ((Elf64_Dyn *)dyns)[dynpos].d_tag = DT_NULL;
+        ((Elf64_Dyn *)dyns)[dynpos].d_un.d_val = 0x0;
+       }
      }
 
    if (lseek(fd, PHDR(p_offset), SEEK_SET) == -1
        || write(fd, dyns, PHDR(p_filesz)) != (int)PHDR(p_filesz))
      {
        perror ("writing dynamic section");
+       free(dyns);
+       elf_close(fd);
        return 1;
      }
 
+   free(dyns);
    elf_close(fd);
 
    return 0;
